/*
FUSE: Filesystem in Userspace
Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>

This program can be distributed under the terms of the GNU GPL.
See the file COPYING.
*/

/** @file
 *
 * minimal example filesystem using low-level API
 *
 * Compile with:
 *
 *     gcc -Wall hello_ll.c `pkg-config fuse3 --cflags --libs` -o hello_ll
 *
 * ## Source code ##
 * \include hello_ll.c
 */

// #define FUSE_USE_VERSION 31
#define FUSE_USE_VERSION 26

#include <fuse_lowlevel.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>
#include <pthread.h>

#define ZERO_FILE_SIZE 0x10000000

#define INODE_FLAG_FILE 0
#define INODE_FLAG_DIR 1
#define INODE_FLAG_DENTRY_CACHE 2
#define INODE_FLAG_ATTR_CACHE 4
#define INODE_FLAG_DIRECT_MODE 8
#define INODE_FLAG_ZERO 16
#define INODE_FLAG_SLOW 32

static struct _hello_name {
	const char *fname;
	uint32_t inode;
	uint32_t flags;
} hello_name_tab[] = {
	{".",		1,	INODE_FLAG_DIR|INODE_FLAG_DENTRY_CACHE|INODE_FLAG_ATTR_CACHE},
	{"..",		1,	INODE_FLAG_DIR|INODE_FLAG_DENTRY_CACHE|INODE_FLAG_ATTR_CACHE},
	{"hello1",	2,	INODE_FLAG_FILE},
	{"hello2",	3,	INODE_FLAG_FILE},
	{"hello_dc1",	4,	INODE_FLAG_FILE|INODE_FLAG_DENTRY_CACHE},
	{"hello_dc2",	5,	INODE_FLAG_FILE|INODE_FLAG_DENTRY_CACHE},
	{"hello_ac1",	6,	INODE_FLAG_FILE|INODE_FLAG_ATTR_CACHE},
	{"hello_ac2",	7,	INODE_FLAG_FILE|INODE_FLAG_ATTR_CACHE},
	{"hello_adc1",	8,	INODE_FLAG_FILE|INODE_FLAG_DENTRY_CACHE|INODE_FLAG_ATTR_CACHE},
	{"hello_adc2",	9,	INODE_FLAG_FILE|INODE_FLAG_DENTRY_CACHE|INODE_FLAG_ATTR_CACHE},
	{"zero_direct",10,	INODE_FLAG_FILE|INODE_FLAG_ZERO|INODE_FLAG_DIRECT_MODE},
	{"zero_cached",11,	INODE_FLAG_FILE|INODE_FLAG_ZERO},
	{"slow_direct",12,	INODE_FLAG_FILE|INODE_FLAG_ZERO|INODE_FLAG_DIRECT_MODE|INODE_FLAG_SLOW},
	{"slow_cached",13,	INODE_FLAG_FILE|INODE_FLAG_ZERO|INODE_FLAG_SLOW},
	{NULL,0,0}};

typedef struct _mem_file {
	uint32_t size;
	uint32_t length;
	time_t atime,mtime,ctime;
	char *databuff;
	pthread_mutex_t lock;
} mem_file;

typedef struct _io_block_size {
	uint32_t bsize;
	uint32_t count;
	struct _io_block_size *next;
} io_block_size;

typedef struct _file_handle {
	pthread_mutex_t lock;
	io_block_size *bsizes;
	uint8_t io_in_progress;
} file_handle;

static mem_file hello_file;
static mem_file zero_file;
static struct fuse_chan *fusech;

static inline void hello_init(struct fuse_chan *ch) {
	time_t now = time(NULL);

	pthread_mutex_init(&hello_file.lock,NULL);
	hello_file.size = 0x10000;
	hello_file.length = 0;
	hello_file.atime = now;
	hello_file.mtime = now;
	hello_file.ctime = now;
	hello_file.databuff = malloc(hello_file.size);
	pthread_mutex_init(&zero_file.lock,NULL);
	zero_file.size = 0x10000;
	zero_file.length = ZERO_FILE_SIZE;
	zero_file.atime = now;
	zero_file.mtime = now;
	zero_file.ctime = now;
	zero_file.databuff = malloc(zero_file.size);
	fusech = ch;
}

static inline void hello_cleanup(void) {
	pthread_mutex_lock(&zero_file.lock);
	pthread_mutex_unlock(&zero_file.lock);
	pthread_mutex_destroy(&zero_file.lock);
	free(zero_file.databuff);
	pthread_mutex_lock(&hello_file.lock);
	pthread_mutex_unlock(&hello_file.lock);
	pthread_mutex_destroy(&hello_file.lock);
	free(hello_file.databuff);
}

static inline int hello_find_by_inode(fuse_ino_t ino) {
	int i;
	for (i=0 ; hello_name_tab[i].fname!=NULL ; i++) {
		if (hello_name_tab[i].inode == ino) {
			return i;
		}
	}
	return -1;
}

static inline int hello_find_by_name(const char *fname) {
	int i;
	for (i=0 ; hello_name_tab[i].fname!=NULL ; i++) {
		if (strcmp(hello_name_tab[i].fname,fname)==0) {
			return i;
		}
	}
	return -1;
}

static inline void hello_add_bsize(file_handle *fh,uint32_t size) {
	io_block_size *iobs;

	for (iobs = fh->bsizes ; iobs!=NULL ; iobs=iobs->next) {
		if (iobs->bsize == size) {
			iobs->count++;
			return;
		}
	}
	iobs = malloc(sizeof(io_block_size));
	iobs->bsize = size;
	iobs->count = 1;
	iobs->next = fh->bsizes;
	fh->bsizes = iobs;
}

static inline void hello_print_and_free_bsizes(file_handle *fh) {
	io_block_size *iobs,*iobsn;

	for (iobs = fh->bsizes ; iobs!=NULL ; iobs=iobsn) {
		iobsn = iobs->next;
		fprintf(stderr,"bsize: %u ; count: %u\n",iobs->bsize,iobs->count);
		free(iobs);
	}
	fh->bsizes = NULL;
}

#if 0
static inline void hello_inval_entries(void) { // not used (doesn't work on FreeBSD)
	int i;
	for (i=0 ; hello_name_tab[i].fname!=NULL ; i++) {
		if ((hello_name_tab[i].flags & (INODE_FLAG_DIR|INODE_FLAG_ZERO))==0) {
			fuse_lowlevel_notify_inval_entry(fusech,1,hello_name_tab[i].fname,strlen(hello_name_tab[i].fname));
		}
	}
}
#endif

#if FUSE_USE_VERSION >= 26
static void hello_ll_statfs(fuse_req_t req,fuse_ino_t ino) {
#else
static void hello_ll_statfs(fuse_req_t req) {
#endif
	struct statvfs stfsbuf;

#if FUSE_USE_VERSION >= 26
        (void)ino;
#endif

	stfsbuf.f_namemax = 255;
	stfsbuf.f_frsize = 0x10000;
	stfsbuf.f_bsize = 0x10000;
	stfsbuf.f_blocks = 1000;
	stfsbuf.f_bfree = 500;
	stfsbuf.f_bavail = 500;
	stfsbuf.f_files = 10;
	stfsbuf.f_ffree = 5;
	stfsbuf.f_favail = 5;
	fuse_reply_statfs(req,&stfsbuf);
}



static void hello_stat(int i, struct stat *stbuf) {

	memset(stbuf, 0, sizeof(struct stat));
	stbuf->st_ino = hello_name_tab[i].inode;
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		stbuf->st_mode = S_IFDIR | 0777;
		stbuf->st_nlink = 2;
	} else {
		stbuf->st_mode = S_IFREG | 0666;
		stbuf->st_nlink = 1;
		if (hello_name_tab[i].flags & INODE_FLAG_ZERO) {
			stbuf->st_size = ZERO_FILE_SIZE;
			pthread_mutex_lock(&zero_file.lock);
			stbuf->st_atime = zero_file.atime;
			stbuf->st_mtime = zero_file.mtime;
			stbuf->st_ctime = zero_file.ctime;
			pthread_mutex_unlock(&zero_file.lock);
		} else {
			pthread_mutex_lock(&hello_file.lock);
			stbuf->st_atime = hello_file.atime;
			stbuf->st_mtime = hello_file.mtime;
			stbuf->st_ctime = hello_file.ctime;
			stbuf->st_size = hello_file.length;
			pthread_mutex_unlock(&hello_file.lock);
		}
	}
}

static void hello_ll_getattr(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
	struct stat repstbuf;
	int i;

	(void) fi;

	i = hello_find_by_inode(ino);
	if (i<0) {
		fuse_reply_err(req, ENOENT);
	} else {
		hello_stat(i, &repstbuf);
		fuse_reply_attr(req, &repstbuf, (hello_name_tab[i].flags&INODE_FLAG_ATTR_CACHE)?1.0:0.0);
	}
}

static void hello_ll_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *stbuf, int to_set, struct fuse_file_info *fi) {
	struct stat repstbuf;
	time_t now;
	int i;

	i = hello_find_by_inode(ino);
	if (i<0) {
		fuse_reply_err(req, ENOENT);
	} else {
		if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
			fuse_reply_err(req, EPERM);
			return;
		}
		now = time(NULL);
		if (to_set & FUSE_SET_ATTR_SIZE) {
//			fprintf(stderr,"inode %llu ; set size to: %llu\n",(unsigned long long)ino,(unsigned long long)(stbuf->st_size));
			if ((hello_name_tab[i].flags & INODE_FLAG_ZERO) == 0) {
				uint32_t newsize = stbuf->st_size;
				pthread_mutex_lock(&hello_file.lock);

				if (newsize > hello_file.size) {
					newsize += 0xFFFF;
					newsize &= 0xFFFF0000;
					hello_file.databuff = realloc(hello_file.databuff,newsize);
					memset(hello_file.databuff+hello_file.size,0,newsize-hello_file.size);
					hello_file.size = newsize;
				}
				hello_file.length = stbuf->st_size;
				hello_file.mtime = now;
				hello_file.ctime = now;
				pthread_mutex_unlock(&hello_file.lock);
			} else {
				pthread_mutex_lock(&zero_file.lock);
				zero_file.mtime = now;
				zero_file.ctime = now;
				pthread_mutex_unlock(&zero_file.lock);
			}
		}
		if (to_set & FUSE_SET_ATTR_ATIME) {
//			fprintf(stderr,"inode %llu ; set atime to: %llu\n",(unsigned long long)ino,(unsigned long long)(stbuf->st_atime));
			if ((hello_name_tab[i].flags & INODE_FLAG_ZERO) == 0) {
				pthread_mutex_lock(&hello_file.lock);
				hello_file.atime = stbuf->st_atime;
				hello_file.ctime = now;
				pthread_mutex_unlock(&hello_file.lock);
			} else {
				pthread_mutex_lock(&zero_file.lock);
				zero_file.atime = stbuf->st_atime;
				zero_file.ctime = now;
				pthread_mutex_unlock(&zero_file.lock);
			}
		}
		if (to_set & FUSE_SET_ATTR_MTIME) {
//			fprintf(stderr,"inode %llu ; set mtime to: %llu\n",(unsigned long long)ino,(unsigned long long)(stbuf->st_mtime));
			if ((hello_name_tab[i].flags & INODE_FLAG_ZERO) == 0) {
				pthread_mutex_lock(&hello_file.lock);
				hello_file.mtime = stbuf->st_mtime;
				hello_file.ctime = now;
				pthread_mutex_unlock(&hello_file.lock);
			} else {
				pthread_mutex_lock(&zero_file.lock);
				zero_file.mtime = stbuf->st_mtime;
				zero_file.ctime = now;
				pthread_mutex_unlock(&zero_file.lock);
			}
		}
		hello_stat(i, &repstbuf);
		fuse_reply_attr(req, &repstbuf, (hello_name_tab[i].flags&INODE_FLAG_ATTR_CACHE)?1.0:0.0);
	}
}

static void hello_ll_lookup(fuse_req_t req, fuse_ino_t parent, const char *name) {
	struct fuse_entry_param e;
	int i;

	if (parent==1) {
		i = hello_find_by_name(name);
		if (i>=0) {
			memset(&e, 0, sizeof(e));
			e.ino = hello_name_tab[i].inode;
			e.attr_timeout = (hello_name_tab[i].flags&2)?1.0:0.0;
			e.entry_timeout = (hello_name_tab[i].flags&1)?1.0:0.0;
			hello_stat(i, &e.attr);
			fuse_reply_entry(req, &e);
			return;
		}
	}
	fuse_reply_err(req, ENOENT);
}

struct dirbuf {
	char *p;
	size_t size;
};

static void dirbuf_add(fuse_req_t req, struct dirbuf *b, const char *name, fuse_ino_t ino) {
	struct stat stbuf;
	size_t oldsize = b->size;
	b->size += fuse_add_direntry(req, NULL, 0, name, NULL, 0);
	b->p = (char *) realloc(b->p, b->size);
	memset(&stbuf, 0, sizeof(stbuf));
	stbuf.st_ino = ino;
	fuse_add_direntry(req, b->p + oldsize, b->size - oldsize, name, &stbuf, b->size);
}

#define min(x, y) ((x) < (y) ? (x) : (y))

static int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize, off_t off, size_t maxsize) {
	if (off < bufsize) {
		return fuse_reply_buf(req, buf + off, min(bufsize - off, maxsize));
	} else {
		return fuse_reply_buf(req, NULL, 0);
	}
}

static void hello_ll_opendir(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
	if (ino != 1) {
		fuse_reply_err(req, ENOTDIR);
	} else {
		fuse_reply_open(req,fi);
	}
}

static void hello_ll_readdir(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off, struct fuse_file_info *fi) {
	(void) fi;
	int i;

	if (ino != 1) {
		fuse_reply_err(req, ENOTDIR);
	} else {
		struct dirbuf b;

		memset(&b, 0, sizeof(b));
		for (i=0 ; hello_name_tab[i].fname!=NULL ; i++) {
			dirbuf_add(req, &b, hello_name_tab[i].fname, hello_name_tab[i].inode);
		}
		reply_buf_limited(req, b.p, b.size, off, size);
		free(b.p);
	}
}

static void hello_ll_releasedir(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
	(void) fi;

	if (ino != 1) {
		fuse_reply_err(req, EBADF);
	} else {
		fuse_reply_err(req,0);
	}
}

static void hello_ll_open(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
	file_handle *fh;
	int i;

	i = hello_find_by_inode(ino);
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		fuse_reply_err(req, EISDIR);
	} else {
		if (hello_name_tab[i].flags & INODE_FLAG_DIRECT_MODE) {
			fi->direct_io = 1;
		}
		fi->keep_cache = 0;
		fh = malloc(sizeof(file_handle));
		pthread_mutex_init(&(fh->lock),NULL);
		fh->bsizes = NULL;
		fh->io_in_progress = 0;
		fi->fh = (unsigned long long int)(void*)fh;
		fuse_reply_open(req, fi);
	}
}

static void hello_ll_read(fuse_req_t req, fuse_ino_t ino, size_t size, off_t off, struct fuse_file_info *fi) {
	file_handle *fh;
	int i;

	i = hello_find_by_inode(ino);
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		fuse_reply_err(req, EBADF);
		return;
	}
	fh = (file_handle*)(fi->fh);
	if (fh==NULL) {
		fuse_reply_err(req, EBADF);
		return;
	}
	pthread_mutex_lock(&(fh->lock));
	if (hello_name_tab[i].flags & INODE_FLAG_ZERO) {
		hello_add_bsize(fh,size);
	}
	if (hello_name_tab[i].flags & INODE_FLAG_SLOW) {
		fh->io_in_progress++;
		pthread_mutex_unlock(&(fh->lock));
		sleep(10);
		pthread_mutex_lock(&(fh->lock));
		fh->io_in_progress--;
	}
	pthread_mutex_unlock(&(fh->lock));
	if (hello_name_tab[i].flags & INODE_FLAG_ZERO) {
		pthread_mutex_lock(&zero_file.lock);
		if (size > zero_file.size) {
			uint32_t newsize;
			newsize = size;
			newsize += 0xFFFF;
			newsize &= 0xFFFF0000;
			zero_file.databuff = realloc(zero_file.databuff,newsize);
			memset(zero_file.databuff,0,newsize);
			zero_file.size = newsize;
		}
		if (off < ZERO_FILE_SIZE) {
			if (off+size < ZERO_FILE_SIZE) {
				fuse_reply_buf(req,zero_file.databuff,size);
			} else {
				fuse_reply_buf(req,zero_file.databuff,ZERO_FILE_SIZE-off);
			}
		} else {
			fuse_reply_buf(req, NULL, 0);
		}
		zero_file.atime = time(NULL);
		pthread_mutex_unlock(&zero_file.lock);
	} else {
		pthread_mutex_lock(&hello_file.lock);
		if (off < hello_file.length) {
			if (off+size < hello_file.length) {
				fuse_reply_buf(req,hello_file.databuff+off,size);
			} else {
				fuse_reply_buf(req,hello_file.databuff+off,hello_file.length-off);
			}
		} else {
			fuse_reply_buf(req, NULL, 0);
		}
		hello_file.atime = time(NULL);
		pthread_mutex_unlock(&hello_file.lock);
	}
}

static void hello_ll_write(fuse_req_t req, fuse_ino_t ino, const char *buf, size_t size, off_t off, struct fuse_file_info *fi) {
	file_handle *fh;
	int i;

	i = hello_find_by_inode(ino);
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		fuse_reply_err(req, EBADF);
		return;
	}
	fh = (file_handle*)(fi->fh);
	if (fh==NULL) {
		fuse_reply_err(req, EBADF);
		return;
	}
	pthread_mutex_lock(&(fh->lock));
	if (hello_name_tab[i].flags & INODE_FLAG_ZERO) {
		hello_add_bsize(fh,size);
	}
	if (hello_name_tab[i].flags & INODE_FLAG_SLOW) {
		fh->io_in_progress++;
		pthread_mutex_unlock(&(fh->lock));
		sleep(10);
		pthread_mutex_lock(&(fh->lock));
		fh->io_in_progress--;
	}
	pthread_mutex_unlock(&(fh->lock));
	if (hello_name_tab[i].flags & INODE_FLAG_ZERO) {
		pthread_mutex_lock(&zero_file.lock);
		zero_file.mtime = time(NULL);
		zero_file.ctime = time(NULL);
		pthread_mutex_unlock(&zero_file.lock);
		fuse_reply_write(req,size);
	} else {
		pthread_mutex_lock(&hello_file.lock);

		if (off+size > hello_file.size) {
			uint32_t newsize = off+size;
			newsize += 0xFFFF;
			newsize &= 0xFFFF0000;
			hello_file.databuff = realloc(hello_file.databuff,newsize);
			memset(hello_file.databuff+hello_file.size,0,newsize-hello_file.size);
			hello_file.size = newsize;
		}
		memcpy(hello_file.databuff+off,buf,size);
		if (off+size > hello_file.length) {
			hello_file.length = off+size;
		}
		hello_file.mtime = time(NULL);
		hello_file.ctime = time(NULL);
		pthread_mutex_unlock(&hello_file.lock);
		fuse_reply_write(req,size);
	}
}

void hello_ll_flush(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
	(void) fi;
	int i;

	i = hello_find_by_inode(ino);
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		fuse_reply_err(req, EBADF);
	} else {
		fuse_reply_err(req,0);
	}
}

void hello_ll_fsync(fuse_req_t req, fuse_ino_t ino, int datasync, struct fuse_file_info *fi) {
	(void) fi;
	(void) datasync;
	int i;

	i = hello_find_by_inode(ino);
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		fuse_reply_err(req, EBADF);
	} else {
		fuse_reply_err(req,0);
	}
}

void hello_ll_release(fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi) {
	file_handle *fh;
	int i;

	i = hello_find_by_inode(ino);
	if (hello_name_tab[i].flags & INODE_FLAG_DIR) {
		fuse_reply_err(req, EBADF);
		return;
	}
	fh = (file_handle*)(fi->fh);
	if (fh==NULL) {
		fuse_reply_err(req, EBADF);
		return;
	}
	pthread_mutex_lock(&(fh->lock));
	hello_print_and_free_bsizes(fh);
	if (fh->io_in_progress>0) {
		fprintf(stderr,"release when there are still unfinished I/O !!!\n");
		// ignore it here, so it will mean memory leak !!!
		// formaly fh should be added to some structure to be released later
		pthread_mutex_unlock(&(fh->lock));
	} else {
		pthread_mutex_unlock(&(fh->lock));
		pthread_mutex_destroy(&(fh->lock));
		free(fh);
	}
	fuse_reply_err(req,0);
}

static struct fuse_lowlevel_ops hello_ll_oper = {
	.statfs		= hello_ll_statfs,	
	.lookup		= hello_ll_lookup,
	.getattr	= hello_ll_getattr,
	.setattr	= hello_ll_setattr,
	.opendir	= hello_ll_opendir,
	.readdir	= hello_ll_readdir,
	.releasedir	= hello_ll_releasedir,
	.open		= hello_ll_open,
	.read		= hello_ll_read,
	.write		= hello_ll_write,
	.flush		= hello_ll_flush,
	.fsync		= hello_ll_fsync,
	.release	= hello_ll_release,
};


int main(int argc, char *argv[]) {
        struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
        struct fuse_chan *ch;
        char *mountpoint;
        int err = -1;

	if (fuse_parse_cmdline(&args, &mountpoint, NULL, NULL) != -1 && (ch = fuse_mount(mountpoint, &args)) != NULL) {
		struct fuse_session *se;

		se = fuse_lowlevel_new(&args, &hello_ll_oper, sizeof(hello_ll_oper), NULL);
		if (se != NULL) {
			if (fuse_set_signal_handlers(se) != -1) {
				fuse_session_add_chan(se, ch);

				hello_init(ch);

				err = fuse_session_loop_mt(se);

				hello_cleanup();

				fuse_remove_signal_handlers(se);
				fuse_session_remove_chan(ch);
			}
			fuse_session_destroy(se);
		}
		fuse_unmount(mountpoint, ch);
	}
	fuse_opt_free_args(&args);

	return err ? 1 : 0;
}

// FUSE 3.1 MAIN:
#if 0
int main(int argc, char *argv[]) {
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
	struct fuse_session *se;
	struct fuse_cmdline_opts opts;
	int ret = -1;

	if (fuse_parse_cmdline(&args, &opts) != 0) {
		return 1;
	}
	if (opts.show_help) {
		printf("usage: %s [options] <mountpoint>\n\n", argv[0]);
		fuse_cmdline_help();
		fuse_lowlevel_help();
		ret = 0;
		goto err_out1;
	} else if (opts.show_version) {
		printf("FUSE library version %s\n", fuse_pkgversion());
		fuse_lowlevel_version();
		ret = 0;
		goto err_out1;
	}

	se = fuse_session_new(&args, &hello_ll_oper, sizeof(hello_ll_oper), NULL);
	if (se == NULL) {
		goto err_out1;
	}

	if (fuse_set_signal_handlers(se) != 0) {
		goto err_out2;
	}

	if (fuse_session_mount(se, opts.mountpoint) != 0) {
		goto err_out3;
	}

	fuse_daemonize(opts.foreground);

	hello_init();

	/* Block until ctrl+c or fusermount -u */
	if (opts.singlethread) {
		ret = fuse_session_loop(se);
	} else {
		ret = fuse_session_loop_mt(se, opts.clone_fd);
	}

	hello_cleanup();

	fuse_session_unmount(se);
err_out3:
	fuse_remove_signal_handlers(se);
err_out2:
	fuse_session_destroy(se);
err_out1:
	free(opts.mountpoint);
	fuse_opt_free_args(&args);

	return ret ? 1 : 0;
}
#endif
